//////////////////////////////////////////////////////////////////////////////
//
//  Copyright (C) 2007-2011 Infinite Z, Inc.  All Rights Reserved.
//
//  File:       HeadTrackingScene.cpp
//  Content:    Sample code demonstrating use of 
//				      ZSStereo's Head Tracking in Default (simple) mode.
//
//////////////////////////////////////////////////////////////////////////////

#include "HeadTrackingScene.h"


#include "ZSpace/Common/Math/Matrix4.h"

using namespace zspace::stereo;
using namespace zspace::tracker;


HeadTrackingScene* gHeadTrackingScene;						//S3D

// Global declarations
GLfloat DIFFUSE_LIGHT[]   = {0.4f, 0.4f, 0.4f, 1.0f};	 /* diffuse light. */
GLfloat LIGHT_POSITION[]  = {0.2f, 0.2f, 0.6f, 1.0}; 
GLfloat LMODEL_AMBIENT[]  = {0.5, 0.5, 0.5, 1.0}; 
GLfloat MAT_AMBIENT[]     = {1.0, 1.0, 1.0, 1.0};

HeadTrackingScene::HeadTrackingScene(int windowX, int windowY, int windowWidth, int windowHeight)
: m_windowX(windowX), 
  m_windowY(windowY), 
  m_windowWidth(windowWidth), 
  m_windowHeight(windowHeight),
  m_stereoViewport(NULL),
  m_stereoWindow(NULL),
  m_headTarget(NULL),
  m_isHeardTrackingEnabled(true)
{
  // <SampleCode2>

  m_trackerSystem = new zspace::tracker::TrackerSystem();
  m_headTarget    = m_trackerSystem->getDefaultHeadTarget();

  // </SampleCode2>
}


HeadTrackingScene::~HeadTrackingScene()
{
  m_trackerSystem->removeReference();							//S3D
}


bool HeadTrackingScene::initialize(HWND hWnd)
{
  // Initialize the OpenGl rendering context.
  initializeGl(hWnd);											//S3D do we need this?

  // Initialize the stereo window and viewport.
  initializeStereo();

  // Load textures/models.
  initializeScene();

  return true;
}


void HeadTrackingScene::clean()
{
  m_stereoViewport->removeReference();							//S3D
  m_stereoWindow->removeReference();							//S3D

  // Make the rendering context not current 
//  wglMakeCurrent(NULL, NULL) ;					//S3D

  // Release the device context 
//  ReleaseDC(m_hWnd, m_hDc) ; 						//S3D

  // Delete the rendering context 
//  wglDeleteContext(m_hGlrc); 						//S3D
}


void HeadTrackingScene::renderFrame()
{
  // <SampleCode3>

  // Get head tracker pose and pass it to the stereo frustum.
  m_trackerSystem->captureTargets();

  // NOTE: Function setHeadTrackingEnabled(false) of stereoFrustum actually discontinues updating head position. 
  // Setting an identity matrix actually bring the view into default head view. 
  if (m_headTarget && m_headTarget->isEnabled() && m_isHeardTrackingEnabled)
  {
    zspace::common::Matrix4 headPose;
    m_headTarget->getPose(headPose);
    m_stereoViewport->getStereoFrustum()->setHeadPose(headPose);
  }
  else
  {
    m_stereoViewport->getStereoFrustum()->setHeadPose(zspace::common::Matrix4::IDENTITY());
  }

  // </SampleCode3>

  // Update the cube rotation
  glMatrixMode(GL_MODELVIEW);
  glRotatef((GLfloat)-0.8, (GLfloat)0.0, (GLfloat)1.0, (GLfloat)0.0);

  // Render each eye view
  renderEye(StereoFrustum::STEREO_EYE_LEFT);
  renderEye(StereoFrustum::STEREO_EYE_RIGHT);

  // Inform the window's stereo buffer that rendering is complete.
  m_stereoWindow->getStereoBuffer()->endStereo();

  // Show the back buffer
  Sleep(16);
  SwapBuffers(m_hDc);
}


void HeadTrackingScene::onMove(int x, int y)
{
  if (m_stereoWindow)													//S3D
    m_stereoWindow->move(x, y);
}


void HeadTrackingScene::onResize(int width, int height)
{
  if (m_stereoWindow) 
  {
    m_stereoWindow->resize(width, height);								//S3D
    glViewport(0, 0, width, height);
  }
}


void HeadTrackingScene::onSwapRenderTarget()
{
  if (m_stereoWindow && m_stereoWindow->getStereoBuffer())				//S3D, swap L/R eyes
    m_stereoWindow->getStereoBuffer()->swapRenderTarget();
}


void HeadTrackingScene::toggleHeadTracking()
{
  m_isHeardTrackingEnabled = !m_isHeardTrackingEnabled;
}


void HeadTrackingScene::initializeGl(HWND hWnd)
{
  m_hWnd = hWnd;

  // Obtain a device context for the window 
  m_hDc = GetDC(m_hWnd);

  // Set an appropriate pixel format
  PIXELFORMATDESCRIPTOR pfd = { 
    sizeof(PIXELFORMATDESCRIPTOR),   // size of this pfd 
    1,                     // version number 
    PFD_DRAW_TO_WINDOW |   // support window 
    PFD_SUPPORT_OPENGL |   // support OpenGL 
    PFD_STEREO         |   // Stereo, Z3 does not need this flag
    PFD_DOUBLEBUFFER,      // double buffered 
    PFD_TYPE_RGBA,         // RGBA type 
    24,                    // 24-bit color depth 
    0, 0, 0, 0, 0, 0,      // color bits ignored 
    0,                     // no alpha buffer 
    0,                     // shift bit ignored 
    0,                     // no accumulation buffer 
    0, 0, 0, 0,            // accum bits ignored 
    32,                    // 32-bit z-buffer 
    0,                     // no stencil buffer 
    0,                     // no auxiliary buffer 
    PFD_MAIN_PLANE,        // main layer 
    0,                     // reserved 
    0, 0, 0                // layer masks ignored 
  }; 

  int  pixelFormat; 

  // Get the best available match of pixel format for the device context  
  pixelFormat = ChoosePixelFormat(m_hDc, &pfd); 

  // Make that the pixel format of the device context 
  SetPixelFormat(m_hDc, pixelFormat, &pfd);

  // Create an OpenGL rendering context 
  m_hGlrc = wglCreateContext(m_hDc);

  // Make it this thread's current rendering context 
  wglMakeCurrent(m_hDc, m_hGlrc);
}


void HeadTrackingScene::initializeStereo()	//S3D does not actually create a window or use any MSW api
{
  // Create the stereo window
  m_stereoWindow = StereoGl::createStereoWindow(m_windowX, m_windowY, m_windowWidth, m_windowHeight);

  // Create the stereo viewport.
  m_stereoViewport = StereoGl::createStereoViewport();
  m_stereoViewport->setUsingWindowSize(true);

  // Add the viewport to the stereo window.
  m_stereoWindow->addStereoViewport(m_stereoViewport);
}


// Load 3D models into the video RAM.
void HeadTrackingScene::initializeScene()
{
  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();

  const zspace::common::DisplayInfo::Display* display = m_stereoWindow->getCurrentDisplay();

  if (display)
  {
    // Adjust the camera based on display position
    float cameraPosition[3];
    cameraPosition[0] = -display->displayOffset[0]; 
    cameraPosition[1] = -display->displayOffset[1];
    cameraPosition[2] = -display->displayOffset[2];

    // TODO:: Further adjustment on display angle to create a horizontal desktop

    // Look at the center of the screen 
    float lookAt[3];
    lookAt[0] = cameraPosition[0] + display->displayOffset[0];
    lookAt[1] = cameraPosition[1] + display->displayOffset[1];
    lookAt[2] = cameraPosition[2] + display->displayOffset[2];

    gluLookAt(cameraPosition[0], cameraPosition[1], cameraPosition[2],
              lookAt[0],         lookAt[1],         lookAt[2],
              0.0f,              1.0f,              0.0f);		        // Up is in positive Y direction.
  }

  // Set material light reflection properties
  glMaterialfv(GL_FRONT, GL_AMBIENT, MAT_AMBIENT);
  glMaterialfv(GL_FRONT, GL_DIFFUSE, MAT_AMBIENT);
  glColorMaterial(GL_FRONT, GL_AMBIENT_AND_DIFFUSE);
  glEnable(GL_COLOR_MATERIAL);

  // Enable a single OpenGL light.
  glLightfv(GL_LIGHT0, GL_POSITION, LIGHT_POSITION);
  glLightfv(GL_LIGHT0, GL_DIFFUSE, DIFFUSE_LIGHT);
  glLightModelfv(GL_LIGHT_MODEL_AMBIENT, LMODEL_AMBIENT);

  // Enable lighting
  glEnable(GL_LIGHTING);
  glEnable(GL_LIGHT0);
  glShadeModel(GL_SMOOTH);

  // Enable depth testing
  glEnable(GL_DEPTH_TEST);

  /* Initialize texture usage */
  glGenTextures(1, &m_textureName);
  glBindTexture(GL_TEXTURE_2D ,m_textureName);
  glPixelStorei(GL_UNPACK_ALIGNMENT, 4);

  // Load texture from bmp file. If load fails, try parent dir.
  if(!createTextureFromBmp("Infinite-Z-Logo-42900.bmp"))
    createTextureFromBmp("..\\Infinite-Z-Logo-42900.bmp");

  // Set up and enable texture mapping
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
  glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
  glEnable(GL_TEXTURE_2D);
}


// Fairly limited bmp texture loader. BMP files must be 
// single-plane 24-bit color format.
bool HeadTrackingScene::createTextureFromBmp(char* filename)
{
  if(!filename)
    return false;

  FILE* fp = NULL;
  fopen_s(&fp, filename,"rb");
  if(!fp)
    return false;

  // Read bmp header
  static const int bmpHeaderSize = 54;
  BYTE header[bmpHeaderSize];
  fread( header, bmpHeaderSize, 1, fp );

  // Make sure we've got a bmp file
  if((header[0] != 'B' || header[1] != 'M'))
    goto fail;

  // Make sure have a V3 header
  int dataOffset = *(int*)&header[10];
  if(dataOffset != bmpHeaderSize)
    goto fail;

  // Only handle single color planes
  short numColorPlanes = *(short*)&header[26];
  if(numColorPlanes != 1)
    goto fail;

  // Only handle 24-bit color
  short bitsPerPixel = *(short*)&header[28];
  if(bitsPerPixel != 24)
    goto fail;

  // Get image dimensions from header
  int imageWidth = *(int*)&header[18];
  int imageHeight = *(int*)&header[22];

  // Bmp is 4-byte aligned, compute row width
  int rowWidth = (((imageWidth*3)+3)>>2)<<2;

  // Make sure we have the expected file size
  int expectedFileSize = bmpHeaderSize + (rowWidth * imageHeight);
  int fileSize = *(int*)&header[2];
  if(fileSize != expectedFileSize)
    goto fail;

  // Allocate buffer
  BYTE* data = (BYTE *)malloc(fileSize);
  if(!data)
    goto fail;

  // Read pixel data from bmp 
  fread(data, fileSize-bmpHeaderSize, 1, fp);

  // Done with file
  fclose(fp);

  // Create a texture from the pixel data
  glTexImage2D(GL_TEXTURE_2D, 0, 4, imageWidth, imageHeight, 0, GL_BGR_EXT, GL_UNSIGNED_BYTE, data);

  // glTexImage2D creates a texture from data[] so we can free it now.
  free(data);

  // Success
  return true;

fail:
  fclose(fp);
  return false;
}


void HeadTrackingScene::renderEye(StereoFrustum::StereoEye eye)
{
  GLfloat monoModelViewGl[16];
  GLfloat viewMatrixGl[16];
  GLfloat projectionMatrixGl[16];

  // Copy the mono model-view matrix.
  glMatrixMode(GL_MODELVIEW);
  glGetFloatv(GL_MODELVIEW_MATRIX, monoModelViewGl);

  // Get the stereo frustum.
  StereoFrustum* stereoFrustum = m_stereoViewport->getStereoFrustum();

  if (stereoFrustum)
  {
    // Modify the model-view matrix for eye. 
    // Note: The view matrix contains eye offset plus any rotation required for off-axis projection.
    stereoFrustum->getViewMatrix(eye, viewMatrixGl);

    glLoadMatrixf(viewMatrixGl);
    glMultMatrixf(monoModelViewGl);

    // Set the projection matrix for eye.
    stereoFrustum->getProjectionMatrix(eye, projectionMatrixGl);

    glMatrixMode(GL_PROJECTION);
    glLoadMatrixf(projectionMatrixGl);
  }

  // Select appropriate render target to render to.
  if (eye == StereoFrustum::STEREO_EYE_LEFT)
    m_stereoWindow->getStereoBuffer()->selectRenderTarget(StereoBuffer::GRAPHICS_BUFFER_BACK_LEFT);
  else
    m_stereoWindow->getStereoBuffer()->selectRenderTarget(StereoBuffer::GRAPHICS_BUFFER_BACK_RIGHT);

  // Render scene with modified modelview and projection matrix
  renderScene(); //Draw object centered at (0, 0, 0)	//stereo 3D, insert here, debug zz

  // Restore mono modelview matrix 
  glMatrixMode(GL_MODELVIEW);
  glLoadMatrixf(monoModelViewGl);
}


void HeadTrackingScene::renderScene(void)
{
  // Clear color and depth buffers
  glClearColor(0.0f, 0.0f, 0.0f, 1.0f);
  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

  // Render the cube
  glBegin(GL_QUADS);			
  const float halfSize = CUBE_WIDTH * 0.5f;
  // Side 1
  glNormal3f(0.0f, 0.0f, -1.0f);
  glTexCoord2d(1.0,0.0); glVertex3f( -halfSize, -halfSize, -halfSize);
  glTexCoord2d(1.0,1.0); glVertex3f( -halfSize,  halfSize, -halfSize);
  glTexCoord2d(0.0,1.0); glVertex3f(  halfSize,  halfSize, -halfSize);        
  glTexCoord2d(0.0,0.0); glVertex3f(  halfSize, -halfSize, -halfSize);		

  // Side 2
  glNormal3f(0.0f, 0.0f, 1.0f);
  glTexCoord2d(0.0,1.0); glVertex3f( -halfSize,  halfSize, halfSize);
  glTexCoord2d(0.0,0.0); glVertex3f( -halfSize, -halfSize, halfSize);        
  glTexCoord2d(1.0,0.0); glVertex3f(  halfSize, -halfSize, halfSize);
  glTexCoord2d(1.0,1.0); glVertex3f(  halfSize,  halfSize, halfSize);

  // Side 3
  glNormal3f(0.0f, 1.0f, 0.0f);
  glTexCoord2d(0.0,1.0); glVertex3f( -halfSize, halfSize, -halfSize);
  glTexCoord2d(0.0,0.0); glVertex3f( -halfSize, halfSize,  halfSize);
  glTexCoord2d(1.0,0.0); glVertex3f(  halfSize, halfSize,  halfSize);        
  glTexCoord2d(1.0,1.0); glVertex3f(  halfSize, halfSize, -halfSize);

  // Side 4
  glNormal3f(0.0f, -1.0f, 0.0f);
  glTexCoord2d(0.0,1.0); glVertex3f( -halfSize, -halfSize,  halfSize);
  glTexCoord2d(0.0,0.0); glVertex3f( -halfSize, -halfSize, -halfSize);
  glTexCoord2d(1.0,0.0); glVertex3f(  halfSize, -halfSize, -halfSize);
  glTexCoord2d(1.0,1.0); glVertex3f(  halfSize, -halfSize,  halfSize);

  // Side 5
  glNormal3f(1.0f, 0.0f, 0.0f);
  glTexCoord2d(1.0,1.0); glVertex3f( halfSize,  halfSize, -halfSize);
  glTexCoord2d(0.0,1.0); glVertex3f( halfSize,  halfSize,  halfSize);        
  glTexCoord2d(0.0,0.0); glVertex3f( halfSize, -halfSize,  halfSize);
  glTexCoord2d(1.0,0.0); glVertex3f( halfSize, -halfSize, -halfSize);

  // Side 6
  glNormal3f(-1.0f, 0.0f, 0.0f);
  glTexCoord2d(0.0,1.0); glVertex3f( -halfSize,  halfSize, -halfSize);
  glTexCoord2d(0.0,0.0); glVertex3f( -halfSize, -halfSize, -halfSize);        
  glTexCoord2d(1.0,0.0); glVertex3f( -halfSize, -halfSize,  halfSize);
  glTexCoord2d(1.0,1.0); glVertex3f( -halfSize,  halfSize,  halfSize);

  glEnd();
}

//S3D
//-----------------------------------------------------------------------------
int npInitIZ()
{
	int startPosX = 0, startPosY = 0;

	HWND hWnd = FindWindow("GLUT", "ANTz - GameMode");			//S3D//

  // set up logging
  // In the main function of your application create the Log singleton to log
  // all messages of severity INFO and above (this includes ERR and WARN).
  zspace::common::Log::create( "HeadTracking.log", zspace::common::Log::INFO );
  ZS_LOG(INFO, "HeadTracking startup.\n" );
  printf("HeadTracking startup.\n");

  // Initialize the stereoGL (for left-right frame synchronization)
  if (0 != zspace::stereo::StereoGl::initialize(hWnd))
  {
    // Failed to synchronize the left and right frame.
    ZS_LOG(ERR, "Failed to synchronize the left and right frame.\n" ); 
	printf("Failed to synchronize the left and right frame.\n");
    zspace::common::Log::close();
    return(-1);	//change this to allow user to specify if they need to swap eyes, debug zz
  }

//change to data->io.gl.width...
  gHeadTrackingScene = new HeadTrackingScene(startPosX, startPosY, WINDOW_WIDTH, WINDOW_HEIGHT);

  // Set up and initialize GL
  if (!gHeadTrackingScene->initialize(hWnd))
   return -1;

  return 0;
}

