//////////////////////////////////////////////////////////////////////////////
//
//  Copyright (C) 2007-2011 Infinite Z, Inc.  All Rights Reserved.
//
//  File:       Scene.cpp
//  Content:    Sample code demonstrating use of 
//				      ZSStereo's Head Tracking in Default (simple) mode.
//
//////////////////////////////////////////////////////////////////////////////

#include "Scene.h"
#include <stdio.h>

// Global declarations
GLfloat MAT_AMBIENT[]     = {1.0, 1.0, 1.0, 1.0};


Scene::Scene(void)
{
}

Scene::~Scene(void)
{
}

// Load 3D models into the video RAM.
void
Scene::initialize()
{

  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();

  // Set material light reflection properties
  glMaterialfv(GL_FRONT, GL_AMBIENT, MAT_AMBIENT);
  glMaterialfv(GL_FRONT, GL_DIFFUSE, MAT_AMBIENT);
  glColorMaterial(GL_FRONT, GL_AMBIENT_AND_DIFFUSE);
  glEnable(GL_COLOR_MATERIAL);


  // Initialize texture usage.
  glGenTextures(1, &m_textureName);
  glBindTexture(GL_TEXTURE_2D ,m_textureName);
  glPixelStorei(GL_UNPACK_ALIGNMENT, 4);

  // Load texture from bmp file. If load fails, try parent dir.
  if (!createTextureFromBmp("Infinite-Z-Logo-42900.bmp"))
    createTextureFromBmp("..\\Infinite-Z-Logo-42900.bmp");

  // Set up and enable texture mapping
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
  glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
  glEnable(GL_TEXTURE_2D);
}

void
Scene::lookAt(float  posX, float  posY, float  posZ,
              float lookX, float lookY, float lookZ,
              float   upX, float   upY, float   upZ)
{
  gluLookAt(posX, posY, posZ, lookX, lookY, lookZ, upX, upY, upZ);
}

// Fairly limited bmp texture loader. BMP files must be 
// single-plane 24-bit color format.
bool
Scene::createTextureFromBmp(char* filename)
{
  if (!filename)
    return false;

  FILE* fileHandle = NULL;
  fopen_s(&fileHandle, filename, "rb");

  if (!fileHandle)
    return false;

  // Read the bmp header.
  static const int bmpHeaderSize = 54;
  BYTE header[bmpHeaderSize];
  fread(header, bmpHeaderSize, 1, fileHandle);

  // Make sure we've got a bmp file.
  if ((header[0] != 'B' || header[1] != 'M'))
    goto fail;

  // Make sure have a V3 header.
  int dataOffset = *(int*)&header[10];
  if (dataOffset != bmpHeaderSize)
    goto fail;

  // Only handle single color planes.
  short numColorPlanes = *(short*)&header[26];
  if (numColorPlanes != 1)
    goto fail;

  // Only handle 24-bit color.
  short bitsPerPixel = *(short*)&header[28];
  if (bitsPerPixel != 24)
    goto fail;

  // Get image dimensions from header.
  int imageWidth  = *(int*)&header[18];
  int imageHeight = *(int*)&header[22];

  // Bmp is 4-byte aligned, compute row width.
  int rowWidth = (((imageWidth*3)+3)>>2)<<2;

  // Make sure we have the expected file size.
  int expectedFileSize = bmpHeaderSize + (rowWidth * imageHeight);
  int fileSize = *(int*)&header[2];
  if (fileSize != expectedFileSize)
    goto fail;

  // Allocate buffer.
  BYTE* data = (BYTE *)malloc(fileSize);
  if(!data)
    goto fail;

  // Read pixel data from bmp.
  fread(data, fileSize - bmpHeaderSize, 1, fileHandle);

  // Done with file.
  fclose(fileHandle);

  // Create a texture from the pixel data.
  glTexImage2D(GL_TEXTURE_2D, 0, 4, imageWidth, imageHeight, 0, GL_BGR_EXT, GL_UNSIGNED_BYTE, data);

  // glTexImage2D creates a texture from data[] so we can free it now.
  free(data);

  // Success.
  return true;

fail:
  fclose(fileHandle);
  return false;
}

void
Scene::update()
{
  // Update the cube rotation.
  glMatrixMode(GL_MODELVIEW);
//  glRotatef((GLfloat)-0.8, (GLfloat)0.0, (GLfloat)1.0, (GLfloat)0.0);
}

void
Scene::render()
{
  // Clear color and depth buffers
  glClearColor(0.0f, 0.0f, 0.0f, 1.0f);
  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

  // Render the cube
  glBegin(GL_QUADS);			
  const float halfSize = 0.03f;
  // Side 1
  glNormal3f(0.0f, 0.0f, -1.0f);
  glTexCoord2d(1.0,0.0); glVertex3f( -halfSize, -halfSize, -halfSize);
  glTexCoord2d(1.0,1.0); glVertex3f( -halfSize,  halfSize, -halfSize);
  glTexCoord2d(0.0,1.0); glVertex3f(  halfSize,  halfSize, -halfSize);        
  glTexCoord2d(0.0,0.0); glVertex3f(  halfSize, -halfSize, -halfSize);		

  // Side 2
  glNormal3f(0.0f, 0.0f, 1.0f);
  glTexCoord2d(0.0,1.0); glVertex3f( -halfSize,  halfSize, halfSize);
  glTexCoord2d(0.0,0.0); glVertex3f( -halfSize, -halfSize, halfSize);        
  glTexCoord2d(1.0,0.0); glVertex3f(  halfSize, -halfSize, halfSize);
  glTexCoord2d(1.0,1.0); glVertex3f(  halfSize,  halfSize, halfSize);

  // Side 3
  glNormal3f(0.0f, 1.0f, 0.0f);
  glTexCoord2d(0.0,1.0); glVertex3f( -halfSize, halfSize, -halfSize);
  glTexCoord2d(0.0,0.0); glVertex3f( -halfSize, halfSize,  halfSize);
  glTexCoord2d(1.0,0.0); glVertex3f(  halfSize, halfSize,  halfSize);        
  glTexCoord2d(1.0,1.0); glVertex3f(  halfSize, halfSize, -halfSize);

  // Side 4
  glNormal3f(0.0f, -1.0f, 0.0f);
  glTexCoord2d(0.0,1.0); glVertex3f( -halfSize, -halfSize,  halfSize);
  glTexCoord2d(0.0,0.0); glVertex3f( -halfSize, -halfSize, -halfSize);
  glTexCoord2d(1.0,0.0); glVertex3f(  halfSize, -halfSize, -halfSize);
  glTexCoord2d(1.0,1.0); glVertex3f(  halfSize, -halfSize,  halfSize);

  // Side 5
  glNormal3f(1.0f, 0.0f, 0.0f);
  glTexCoord2d(1.0,1.0); glVertex3f( halfSize,  halfSize, -halfSize);
  glTexCoord2d(0.0,1.0); glVertex3f( halfSize,  halfSize,  halfSize);        
  glTexCoord2d(0.0,0.0); glVertex3f( halfSize, -halfSize,  halfSize);
  glTexCoord2d(1.0,0.0); glVertex3f( halfSize, -halfSize, -halfSize);

  // Side 6
  glNormal3f(-1.0f, 0.0f, 0.0f);
  glTexCoord2d(0.0,1.0); glVertex3f( -halfSize,  halfSize, -halfSize);
  glTexCoord2d(0.0,0.0); glVertex3f( -halfSize, -halfSize, -halfSize);        
  glTexCoord2d(1.0,0.0); glVertex3f( -halfSize, -halfSize,  halfSize);
  glTexCoord2d(1.0,1.0); glVertex3f( -halfSize,  halfSize,  halfSize);

  glEnd();
}